@@ -128,6 +128,20 @@ class CosineDecayScheduler(LearningRateScheduler):
         return self.lr * decayed
 
 
+@export
+class LinearDecayScheduler(LearningRateScheduler):
+    def __init__(self, decay_steps=1000, alpha=0.0, **kwargs):
+        super().__init__(**kwargs)
+        self.decay_steps = decay_steps
+        self.alpha = alpha
+
+    def __call__(self, global_step):
+        global_step = min(global_step, self.decay_steps)
+        # Linear interpolation
+        scaled_lr = self.lr * (1.0 - self.alpha) * (1.0 - global_step / self.decay_steps) + (self.alpha * self.lr)
+        return scaled_lr
+
+
 @export
 class InverseTimeDecayScheduler(LearningRateScheduler):
     def __init__(self, decay_steps=16000, decay_rate=0.05, staircase=False, **kwargs):
@@ -156,12 +170,14 @@ class ExponentialDecayScheduler(LearningRateScheduler):
 
 @export
 class CompositeLRScheduler(LearningRateScheduler):
-    def __init__(self, warm=None, rest=None, **kwargs):
+    def __init__(self, warm=None, rest=None, plateau_steps=0, **kwargs):
         super().__init__(**kwargs)
         self.warm = warm
         self.rest = rest
+        self.plateau_steps = plateau_steps
 
     def __call__(self, global_step):
-        if global_step < self.warm.warmup_steps:
+        total_steps_lr1 = self.warm.warmup_steps + self.plateau_steps
+        if global_step < total_steps_lr1:
             return self.warm(global_step)
-        return self.rest(global_step - self.warm.warmup_steps)
+        return self.rest(global_step - total_steps_lr1)
