@@ -2,7 +2,16 @@ import torch
 import torch.nn as nn
 import numpy as np
 from typing import Dict, List
-from eight_mile.pytorch.layers import Dense, TransformerEncoderStack, TransformerEncoder, TransformerDecoderStack, TransformerDecoder, EmbeddingsStack
+from eight_mile.pytorch.layers import (
+    Dense,
+    TransformerEncoderStack,
+    TransformerEncoder,
+    TransformerDecoderStack,
+    TransformerDecoder,
+    EmbeddingsStack,
+    WithDropout,
+    SingleHeadReduction,
+)
 from eight_mile.pytorch.embeddings import LookupTableEmbeddings, LearnedPositionalLookupTableEmbeddingsWithBias
 
 # BERT HuggingFace Tokenizers checkpoints can be converted into MEAD Baseline Transformer checkpoints
@@ -27,10 +36,10 @@ BERT_HF_LAYER_MAP = {
     ## LN weights
     # The names in of layer norm our transformers are a bit unspecific
     # think of ln1 as ln_x and ln2 as ln_attn_output
-    'bert.encoder.layer.{}.output.LayerNorm.bias': 'generator.encoders.{}.ln1.bias',
-    'bert.encoder.layer.{}.output.LayerNorm.weight': 'generator.encoders.{}.ln1.weight',
-    'bert.encoder.layer.{}.attention.output.LayerNorm.bias': 'generator.encoders.{}.ln2.bias',
-    'bert.encoder.layer.{}.attention.output.LayerNorm.weight': 'generator.encoders.{}.ln2.weight'
+    'bert.encoder.layer.{}.output.LayerNorm.beta': 'generator.encoders.{}.ln1.bias',
+    'bert.encoder.layer.{}.output.LayerNorm.gamma': 'generator.encoders.{}.ln1.weight',
+    'bert.encoder.layer.{}.attention.output.LayerNorm.beta': 'generator.encoders.{}.ln2.bias',
+    'bert.encoder.layer.{}.attention.output.LayerNorm.gamma': 'generator.encoders.{}.ln2.weight'
 }
 
 BERT_HF_FT_LAYER_MAP = {
@@ -53,10 +62,10 @@ BERT_HF_FT_LAYER_MAP = {
     ## LN weights
     # The names in of layer norm our transformers are a bit unspecific
     # think of ln1 as ln_x and ln2 as ln_attn_output
-    'bert.encoder.layer.{}.output.LayerNorm.bias': 'transformer.encoders.{}.ln1.bias',
+    'bert.encoder.layer.{}.output.LayerNorm.beta': 'transformer.encoders.{}.ln1.bias',
     'bert.encoder.layer.{}.output.LayerNorm.weight': 'transformer.encoders.{}.ln1.weight',
     'bert.encoder.layer.{}.attention.output.LayerNorm.beta': 'transformer.encoders.{}.ln2.bias',
-    'bert.encoder.layer.{}.attention.output.LayerNorm.weight': 'transformer.encoders.{}.ln2.weight'
+    'bert.encoder.layer.{}.attention.output.LayerNorm.gamma': 'transformer.encoders.{}.ln2.weight'
 }
 
 BERT_HF_EMBED_MAP = {
@@ -64,8 +73,69 @@ BERT_HF_EMBED_MAP = {
     'bert.embeddings.word_embeddings.weight': 'embeddings.embeddings.0.embeddings.weight',
     'bert.embeddings.position_embeddings.weight': 'embeddings.embeddings.0.pos_embeddings.weight',
     'bert.embeddings.token_type_embeddings.weight': 'embeddings.embeddings.1.embeddings.weight',
-    'bert.embeddings.LayerNorm.bias': 'embeddings.reduction.ln.bias',
-    'bert.embeddings.LayerNorm.weight': 'embeddings.reduction.ln.weight',
+    'bert.embeddings.LayerNorm.beta': 'embeddings.reduction.ln.bias',
+    'bert.embeddings.LayerNorm.gamma': 'embeddings.reduction.ln.weight',
+}
+
+ROBERTA_HF_LAYER_MAP = {
+    ## FFN weights
+    'roberta.encoder.layer.{}.intermediate.dense.weight': 'generator.encoders.{}.ffn.0.layer.weight',
+    'roberta.encoder.layer.{}.intermediate.dense.bias': 'generator.encoders.{}.ffn.0.layer.bias',
+    'roberta.encoder.layer.{}.output.dense.weight': 'generator.encoders.{}.ffn.3.layer.weight',
+    'roberta.encoder.layer.{}.output.dense.bias': 'generator.encoders.{}.ffn.3.layer.bias',
+
+    ## MHA weights
+    'roberta.encoder.layer.{}.attention.self.key.weight': 'generator.encoders.{}.self_attn.w_K.layer.weight',
+    'roberta.encoder.layer.{}.attention.self.key.bias': 'generator.encoders.{}.self_attn.w_K.layer.bias',
+    'roberta.encoder.layer.{}.attention.self.query.weight': 'generator.encoders.{}.self_attn.w_Q.layer.weight',
+    'roberta.encoder.layer.{}.attention.self.query.bias': 'generator.encoders.{}.self_attn.w_Q.layer.bias',
+    'roberta.encoder.layer.{}.attention.self.value.weight': 'generator.encoders.{}.self_attn.w_V.layer.weight',
+    'roberta.encoder.layer.{}.attention.self.value.bias': 'generator.encoders.{}.self_attn.w_V.layer.bias',
+    'roberta.encoder.layer.{}.attention.output.dense.weight': 'generator.encoders.{}.self_attn.w_O.layer.weight',
+    'roberta.encoder.layer.{}.attention.output.dense.bias': 'generator.encoders.{}.self_attn.w_O.layer.bias',
+
+    ## LN weights
+    # The names in of layer norm our transformers are a bit unspecific
+    # think of ln1 as ln_x and ln2 as ln_attn_output
+    'roberta.encoder.layer.{}.output.LayerNorm.bias': 'generator.encoders.{}.ln1.bias',
+    'roberta.encoder.layer.{}.output.LayerNorm.weight': 'generator.encoders.{}.ln1.weight',
+    'roberta.encoder.layer.{}.attention.output.LayerNorm.bias': 'generator.encoders.{}.ln2.bias',
+    'roberta.encoder.layer.{}.attention.output.LayerNorm.weight': 'generator.encoders.{}.ln2.weight'
+}
+
+ROBERTA_HF_FT_LAYER_MAP = {
+    ## FFN weights
+    'roberta.encoder.layer.{}.intermediate.dense.weight': 'transformer.encoders.{}.ffn.0.layer.weight',
+    'roberta.encoder.layer.{}.intermediate.dense.bias': 'transformer.encoders.{}.ffn.0.layer.bias',
+    'roberta.encoder.layer.{}.output.dense.weight': 'transformer.encoders.{}.ffn.3.layer.weight',
+    'roberta.encoder.layer.{}.output.dense.bias': 'transformer.encoders.{}.ffn.3.layer.bias',
+
+    ## MHA weights
+    'roberta.encoder.layer.{}.attention.self.key.weight': 'transformer.encoders.{}.self_attn.w_K.layer.weight',
+    'roberta.encoder.layer.{}.attention.self.key.bias': 'transformer.encoders.{}.self_attn.w_K.layer.bias',
+    'roberta.encoder.layer.{}.attention.self.query.weight': 'transformer.encoders.{}.self_attn.w_Q.layer.weight',
+    'roberta.encoder.layer.{}.attention.self.query.bias': 'transformer.encoders.{}.self_attn.w_Q.layer.bias',
+    'roberta.encoder.layer.{}.attention.self.value.weight': 'transformer.encoders.{}.self_attn.w_V.layer.weight',
+    'roberta.encoder.layer.{}.attention.self.value.bias': 'transformer.encoders.{}.self_attn.w_V.layer.bias',
+    'roberta.encoder.layer.{}.attention.output.dense.weight': 'transformer.encoders.{}.self_attn.w_O.layer.weight',
+    'roberta.encoder.layer.{}.attention.output.dense.bias': 'transformer.encoders.{}.self_attn.w_O.layer.bias',
+
+    ## LN weights
+    # The names in of layer norm our transformers are a bit unspecific
+    # think of ln1 as ln_x and ln2 as ln_attn_output
+    'roberta.encoder.layer.{}.output.LayerNorm.bias': 'transformer.encoders.{}.ln1.bias',
+    'roberta.encoder.layer.{}.output.LayerNorm.weight': 'transformer.encoders.{}.ln1.weight',
+    'roberta.encoder.layer.{}.attention.output.LayerNorm.beta': 'transformer.encoders.{}.ln2.bias',
+    'roberta.encoder.layer.{}.attention.output.LayerNorm.weight': 'transformer.encoders.{}.ln2.weight'
+}
+
+ROBERTA_HF_EMBED_MAP = {
+    ## Embedding weights
+    'roberta.embeddings.word_embeddings.weight': 'embeddings.embeddings.0.embeddings.weight',
+    'roberta.embeddings.position_embeddings.weight': 'embeddings.embeddings.0.pos_embeddings.weight',
+    'roberta.embeddings.token_type_embeddings.weight': 'embeddings.embeddings.1.embeddings.weight',
+    'roberta.embeddings.LayerNorm.bias': 'embeddings.reduction.ln.bias',
+    'roberta.embeddings.LayerNorm.weight': 'embeddings.reduction.ln.weight',
 }
 
 
@@ -82,19 +152,20 @@ def convert_transformers_keys(num_layers: int, d: Dict, nested_layer_map: Dict =
     return m
 
 
-def tlm_load_state_dict(module: nn.Module, checkpoint_file: str):
+def tlm_load_state_dict(module: nn.Module, checkpoint_file: str, map_location=None, str_map = None):
     """
 
     :param tlm: Safely loads the state dict for a transformer encoder
     :param checkpoint_file: The file name
     :return: None
     """
-    str_map = {'transformer': 'generator'}
-    if hasattr(module, 'transformer'):
-        str_map = {'generator': 'transformer'}
+    if str_map is None:
+        str_map = {'transformer': 'generator'}
+        if hasattr(module, 'transformer'):
+            str_map = {'generator': 'transformer'}
     if hasattr(module, 'reduction_layer'):
         str_map.update({'reduction_layer_1': 'reduction_layer'})
-    ckpt_dict = torch.load(checkpoint_file)
+    ckpt_dict = torch.load(checkpoint_file, map_location=map_location)
     renamed = {}
     for k, v in ckpt_dict.items():
         for from_str, to_str in str_map.items():
@@ -113,6 +184,8 @@ def to_weight_array(pytorch_layer: nn.Module, name: str) -> Dict:
     :param name: The name of this layer to serialize
     :return: A Dictionary containing `weights` and `bias` keys
     """
+    if isinstance(pytorch_layer, WithDropout):
+        pytorch_layer = pytorch_layer.layer
     if isinstance(pytorch_layer, Dense):
         pytorch_layer = pytorch_layer.layer
     weights = pytorch_layer.weight.cpu().detach().numpy()
@@ -250,7 +323,7 @@ def from_decoder_array(pytorch_decoder: TransformerDecoder, d: Dict, name: str):
     """
     from_weight_array(pytorch_decoder.ln1, d, f"{name}/ln1")
     from_weight_array(pytorch_decoder.ln2, d, f"{name}/ln2")
-    from_weight_array(pytorch_decoder.ln2, d, f"{name}/ln3")
+    from_weight_array(pytorch_decoder.ln3, d, f"{name}/ln3")
     from_attn_array(pytorch_decoder.src_attn, d, f"{name}/src_attn")
     from_attn_array(pytorch_decoder.self_attn, d, f"{name}/self_attn")
     from_ffn_array(pytorch_decoder.ffn, d, f"{name}/ffn")
@@ -266,7 +339,7 @@ def to_decoder_array(pytorch_decoder: TransformerDecoder, name: str) -> Dict:
     d = {}
     d.update(to_weight_array(pytorch_decoder.ln1, f"{name}/ln1"))
     d.update(to_weight_array(pytorch_decoder.ln2, f"{name}/ln2"))
-    d.update(to_weight_array(pytorch_decoder.ln2, f"{name}/ln3"))
+    d.update(to_weight_array(pytorch_decoder.ln3, f"{name}/ln3"))
     d.update(to_attn_array(pytorch_decoder.self_attn, f"{name}/self_attn"))
     d.update(to_attn_array(pytorch_decoder.src_attn, f"{name}/src_attn"))
     d.update(to_ffn_array(pytorch_decoder.ffn, f"{name}/ffn"))
@@ -302,7 +375,8 @@ def from_embed_array(pytorch_embed: nn.Module, d: Dict, name: str):
     device = pytorch_embed.embeddings.weight.device
     pytorch_embed.embeddings.weight = torch.nn.Parameter(torch.from_numpy(d[f"{name}/weights"]).to(device=device), requires_grad=True)
     if hasattr(pytorch_embed, 'pos_embeddings'):
-        pytorch_embed.pos_embeddings.weight = torch.nn.Parameter(torch.from_numpy(d[f"{name}/pos_weights"]).to(device=device),
+        pos_weights = torch.from_numpy(d[f"{name}/pos_weights"])
+        pytorch_embed.pos_embeddings.weight = torch.nn.Parameter(pos_weights.to(device=device),
                                                                  requires_grad=True)
 
 
@@ -350,6 +424,149 @@ def save_tlm_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] =
     np.savez(npz, **d)
 
 
+def save_tlm_output_npz(tlm: nn.Module, npz: str, embeddings_keys: List[str] = None, name: str = "TLM", verbose: bool = False):
+    """Save a TLM to an NPZ file with an output layer
+
+    :param tlm: A Transformer LM-type module
+    :param npz: A file to save
+    :param embeddings_keys: A key to get embeddings from.  Defaults to `None`, in which case, all embeddings are written
+    :param name: A name for this TLM
+    :param verbose: whether output
+    :return: None
+    """
+    d = to_tlm_array(tlm, embeddings_keys, name)
+    if hasattr(tlm, 'output'):
+        d.update(to_weight_array(tlm.output, name=f'{name}/output'))
+    elif hasattr(tlm, 'output_layer'):
+        d.update(to_weight_array(tlm.output_layer, name=f'{name}/output'))
+    else:
+        raise Exception("No output layer was found")
+    if verbose:
+        print(d.keys())
+    np.savez(npz, **d)
+
+
+
+def to_attn_pool_array(pyt_attn_pool: nn.Module, name: str) -> Dict:
+    """Convert a self-attention module to a set of arrays
+
+    :param pyt_attn_pool: The attention pooling layer of a dual encoder, this could be single head or 2-headed
+    :return: A Dict containing the arrays by key
+    """
+    d = {}
+
+    if isinstance(pyt_attn_pool, SingleHeadReduction):
+        d.update(to_weight_array(pyt_attn_pool.w_Q, f"{name}/w_Q"))
+        d.update(to_weight_array(pyt_attn_pool.w_K, f"{name}/w_K"))
+    else:
+        d.update(to_weight_array(pyt_attn_pool.reduction1.w_Q, f"{name}/reduction1/w_Q"))
+        d.update(to_weight_array(pyt_attn_pool.reduction1.w_K, f"{name}/reduction1/w_K"))
+
+        d.update(to_weight_array(pyt_attn_pool.reduction1.w_Q, f"{name}/reduction2/w_Q"))
+        d.update(to_weight_array(pyt_attn_pool.reduction1.w_K, f"{name}/reduction2/w_K"))
+    return d
+
+
+def from_attn_pool_array(pyt_attn_pool: nn.Module, d: Dict, name: str):
+    """Restore a self-attention pooling module from a set of keys
+
+    :param pyt_attn_pool: The attention pooling layer of a dual encoder, this could be single head or 2-headed
+    :param d: A Dict of arrays by key
+    :param name: The name of the layer
+    """
+
+    if isinstance(pyt_attn_pool, SingleHeadReduction):
+        from_weight_array(pyt_attn_pool.w_Q, d, f"{name}/w_Q")
+        from_weight_array(pyt_attn_pool.w_K, d, f"{name}/w_K")
+
+    else:
+        from_weight_array(pyt_attn_pool.reduction1.w_Q, d, f"{name}/reduction1/w_Q")
+        from_weight_array(pyt_attn_pool.reduction1.w_K, d, f"{name}/reduction1/w_K")
+        from_weight_array(pyt_attn_pool.reduction2.w_Q, d, f"{name}/reduction2/w_Q")
+        from_weight_array(pyt_attn_pool.reduction2.w_K, d, f"{name}/reduction2/w_K")
+
+
+def save_transformer_de_npz(pyt_de: nn.Module, npz: str, embeddings_keys: List[str] = None,
+                            name: str = "TLM", verbose: bool = False):
+    """Save a Transformer de file out
+
+    A Dual-Encoder will have 2 transformer layers with shared weights.  Because of this, when we save we only
+    want to save the first layer.  However, the upper "stacking" layers may be different
+
+    The encoder will have a transformer stack followed optionally by single or dual headed attention, and finally
+    either a linear or FFN stack of layers
+
+    :param pyt_de: A Transformer Dual Encoder module
+    """
+
+    enc = {}
+    transformer = pyt_de.transformer
+    enc.update(to_encoder_stack_array(transformer, name=f"{name}/TransformerEncoderStack"))
+    enc_keys_to_write = embeddings_keys if embeddings_keys else list(pyt_de.embeddings.keys())
+
+    for embeddings_key in enc_keys_to_write:
+        enc.update(to_embed_array(pyt_de.embeddings[embeddings_key], name=f"{name}/Embeddings/{embeddings_key}"))
+
+    enc.update(to_attn_pool_array(pyt_de.reduction_layer, f"{name}/ReductionLayer"))
+
+    ff1 = pyt_de.ff1
+    if isinstance(ff1, nn.Linear):
+        enc.update(to_weight_array(ff1, f"{name}/ff1"))
+    elif not isinstance(ff1, nn.Identity):
+        raise Exception("We dont currently support stacking layers in dual-encoder serialization")
+    ff2 = pyt_de.ff2
+    if isinstance(ff2, nn.Linear):
+        enc.update(to_weight_array(ff2, f"{name}/ff2"))
+    elif not isinstance(ff2, nn.Identity):
+        raise Exception("We dont currently support stacking layers in dual-encoder serialization")
+
+    np.savez(npz, **enc)
+
+
+def load_transformer_de_npz(pyt: nn.Module,
+                            npz: str, embeddings_keys: List[str] = None,
+                            name: str = "TLM"):
+    """Load a dual-encoder from NPZ
+
+    A Dual-Encoder will have 2 transformer layers with shared weights.  Because of this, when we save we only
+    want to save the first layer.  However, the upper "stacking" layers may be different
+
+    The encoder will have a transformer stack followed optionally by single or dual headed attention, and finally
+    either a linear or FFN stack of layers
+
+    :param pyt: A Transformer Dual Encoder module
+    :param npz:
+    :param embeddings_keys:
+    :param name:
+    :return:
+    """
+
+    d = np.load(npz)
+
+    transformer = pyt.transformer
+    from_encoder_stack_array(transformer, d, name=f"{name}/TransformerEncoderStack")
+
+    enc_keys_to_restore = embeddings_keys if embeddings_keys else list(pyt.embeddings.keys())
+
+    for embeddings_key in enc_keys_to_restore:
+        from_embed_array(pyt.embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}")
+
+    from_attn_pool_array(pyt.reduction_layer, d, name=f"{name}/ReductionLayer")
+
+    if hasattr(pyt, 'ff1'):
+        ff1 = pyt.ff1
+        if isinstance(ff1, nn.Linear):
+            from_weight_array(ff1, d, f"{name}/ff1")
+        elif not isinstance(ff1, nn.Identity):
+            raise Exception("We dont currently support stacking layers in dual-encoder serialization")
+    if hasattr(pyt, 'ff2'):
+        ff2 = pyt.ff2
+        if isinstance(ff2, nn.Linear):
+            from_weight_array(ff2, d, f"{name}/ff2")
+        elif not isinstance(ff2, nn.Identity):
+            raise Exception("We dont currently support stacking layers in dual-encoder serialization")
+
+
 def save_transformer_seq2seq_npz(pytorch_seq2seq: nn.Module, npz: str, src_embeddings_keys: List[str] = None,
                                  tgt_embedding_key: str = 'y', name: str = "Seq2Seq", verbose: bool = False):
     """Save a Transformer seq2seq file out
@@ -373,7 +590,7 @@ def save_transformer_seq2seq_npz(pytorch_seq2seq: nn.Module, npz: str, src_embed
     transformer_decoder = pytorch_seq2seq.decoder.transformer_decoder
 
     dec.update(to_decoder_stack_array(transformer_decoder, name=f"{name}/TransformerDecoderStack"))
-    dec.update(to_embed_array(pytorch_seq2seq.tgt_embedding, name=f"{name}/TgtEmbedding/{tgt_embedding_key}"))
+    dec.update(to_embed_array(pytorch_seq2seq.decoder.tgt_embeddings, name=f"{name}/TgtEmbedding/{tgt_embedding_key}"))
 
     if verbose:
         print(enc.keys())
@@ -489,6 +706,24 @@ def load_tlm_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] =
     d = np.load(npz)
     from_tlm_array(pytorch_tlm, d, embeddings_keys, name)
 
+def load_tlm_output_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] = None, name: str = "TLM"):
+    """Restore a TLM-like model (possibly a `nn.Module` for fine-tuning
+
+    We just populate the `TransformerEncoderStack` and the embeddings from weights, all other values remain
+    uninitialized
+
+    :param pytorch_tlm: A TLM-like model
+    :param npz: A file to restore the weights from
+    :param embeddings_key: Name of embeddings to restore, defaults to `None` in which case we restore all embeddings
+    :param name: A name for this primitive
+    :return:
+    """
+    d = np.load(npz)
+    from_tlm_array(pytorch_tlm, d, embeddings_keys, name)
+    if hasattr(pytorch_tlm, 'output_layer'):
+        from_weight_array(pytorch_tlm.output_layer, d, f"{name}/output")
+    else:
+        from_weight_array(pytorch_tlm.output, d, f"{name}/output")
 
 def load_transformer_seq2seq_npz(pytorch_seq2seq: nn.Module, npz: str, src_embeddings_keys: List[str] = None,
                                  tgt_embedding_key: str = 'y', name: str = "Seq2Seq"):
