@@ -161,6 +161,7 @@ class ConstantSchedulerTensorFlow2(LearningRateScheduler, tf.keras.optimizers.sc
         super().__init__(**kwargs)
 
     def __call__(self, global_step):
+        tf.summary.scalar(name='lr', data=self.lr, step=tf.cast(global_step, tf.int64))
         return self.lr
 
     def __str__(self):
@@ -176,7 +177,9 @@ class WarmupLinearSchedulerTensorFlow2(
         super().__init__(warmup_steps=warmup_steps, **kwargs)
 
     def __call__(self, global_step):
-        return tf.minimum(1.0, global_step / float(self.warmup_steps)) * self.lr
+        new_lr = tf.minimum(1.0, global_step / float(self.warmup_steps)) * self.lr
+        tf.summary.scalar(name='warmup_lr', data=new_lr, step=tf.cast(global_step, tf.int64))
+        return new_lr
 
     def __str__(self):
         return type(self).__name__ + "()"
@@ -195,7 +198,9 @@ class CyclicLRSchedulerTensorFlow2(LearningRateScheduler, tf.keras.optimizers.sc
         cycle = tf.floor(1.0 + gs_f / (2.0 * self.decay_steps))
         x = tf.abs(gs_f / self.decay_steps - 2.0 * cycle + 1.0)
         clr = self.lr + (self.max_lr - self.lr) * tf.maximum(0.0, 1.0 - x)
-        return tf.identity(clr, name="lr")
+        new_lr = tf.identity(clr, name="lr")
+        tf.summary.scalar(name='clr_lr', data=new_lr, step=tf.cast(global_step, tf.int64))
+        return new_lr
 
     def __str__(self):
         return type(self).__name__ + "()"
@@ -207,12 +212,14 @@ class SGDRSchedulerTensorFlow2(LearningRateScheduler, tf.keras.optimizers.schedu
         self.first_decay_steps = first_decay_steps
 
     def __call__(self, global_step):
-        return tf.identity(
+        new_lr = tf.identity(
             tf.compat.v1.train.cosine_decay_restarts(
                 self.lr, global_step, first_decay_steps=self.first_decay_steps
             ),
             name="lr",
         )
+        tf.summary.scalar(name='sgdr_lr', data=new_lr, step=tf.cast(global_step, tf.int64))
+        return new_lr
 
     def __str__(self):
         return type(self).__name__ + "()"
@@ -228,24 +235,26 @@ class ZarembaDecaySchedulerTensorFlow2(tf.keras.optimizers.schedules.PiecewiseCo
 
 
 class CompositeLRSchedulerTensorFlow2(tf.keras.optimizers.schedules.LearningRateSchedule):
-    def __init__(self, warm=None, rest=None, **kwargs):
+    def __init__(self, warm=None, rest=None, plateau_steps=0, **kwargs):
         self.warm = warm
         self.rest = rest
+        self.plateau_steps = plateau_steps
 
     def __call__(self, global_step):
+        total_steps_lr1 = self.warm.warmup_steps + self.plateau_steps
         warm_tensor = self.warm(global_step)
 
         def call_warm():
             return warm_tensor
 
-        rest_step = global_step - self.warm.warmup_steps
+        rest_step = global_step - total_steps_lr1
         rest_tensor = self.rest(rest_step)
 
         def call_rest():
             return rest_tensor
 
-        new_lr = tf.cond(global_step < self.warm.warmup_steps, call_warm, call_rest)
-        tf.summary.scalar(name='lr', data=new_lr, step=tf.cast(global_step, tf.int64))
+        new_lr = tf.cond(global_step < total_steps_lr1, call_warm, call_rest)
+        tf.summary.scalar(name='composite_lr', data=new_lr, step=tf.cast(global_step, tf.int64))
         return new_lr
 
     def __str__(self):
@@ -256,18 +265,33 @@ class PiecewiseDecaySchedulerTensorFlow2(tf.keras.optimizers.schedules.Piecewise
     def __init__(self, boundaries, values, **kwargs):
         super().__init__(boundaries, values)
 
+    def __call__(self, global_step):
+        new_lr = super().__call__(global_step)
+        tf.summary.scalar(name='piecewise_decay_lr', data=new_lr, step=tf.cast(global_step, tf.int64))
+        return new_lr
+
 
 class InverseTimeDecaySchedulerTensorFlow2(tf.keras.optimizers.schedules.InverseTimeDecay):
     def __init__(self, decay_steps=16000, decay_rate=0.05, staircase=False, **kwargs):
         lr = kwargs.get("lr", kwargs.get("eta", 0.01))
         super().__init__(lr, decay_steps, decay_rate, staircase, kwargs.get("name"))
 
+    def __call__(self, global_step):
+        new_lr = super().__call__(global_step)
+        tf.summary.scalar(name='inv_time_decay_lr', data=new_lr, step=tf.cast(global_step, tf.int64))
+        return new_lr
+
 
 class ExponentialDecaySchedulerTensorFlow2(tf.keras.optimizers.schedules.ExponentialDecay):
     def __init__(self, decay_steps=16000, decay_rate=0.5, staircase=False, **kwargs):
         lr = kwargs.get("lr", kwargs.get("eta", 0.01))
         super().__init__(lr, decay_steps, decay_rate, staircase, kwargs.get("name"))
 
+    def __call__(self, global_step):
+        new_lr = super().__call__(global_step)
+        tf.summary.scalar(name='exponential_decay_lr', data=new_lr, step=tf.cast(global_step, tf.int64))
+        return new_lr
+
 
 class CosineDecaySchedulerTensorFlow(CosineDecayScheduler, tf.keras.optimizers.schedules.LearningRateSchedule):
     def __init__(self, decay_steps=16000, alpha=0.0, **kwargs):
@@ -278,7 +302,22 @@ class CosineDecaySchedulerTensorFlow(CosineDecayScheduler, tf.keras.optimizers.s
         global_step = tf.math.minimum(global_step, self.decay_steps)
         cosine_decay = 0.5 * (1.0 + tf.cos(3.14159265 * global_step / self.decay_steps))
         decayed = (1 - self.alpha) * cosine_decay + self.alpha
-        return self.lr * decayed
+        new_lr = self.lr * decayed
+        tf.summary.scalar(name='cosine_decay_lr', data=new_lr, step=tf.cast(global_step, tf.int64))
+        return new_lr
+
+
+class LinearDecaySchedulerTensorFlow(LearningRateScheduler, tf.keras.optimizers.schedules.LearningRateSchedule):
+    def __init__(self, decay_steps=1000, alpha=0.0, **kwargs):
+        super().__init__(**kwargs)
+        self.decay_steps = decay_steps
+        self.alpha = alpha
+
+    def __call__(self, global_step):
+        global_step = min(global_step, self.decay_steps)
+        # Linear interpolation
+        scaled_lr = self.lr * (1.0 - self.alpha) * (1.0 - global_step / self.decay_steps) + (self.alpha * self.lr)
+        return scaled_lr
 
 
 if not tf.executing_eagerly():
@@ -302,6 +341,7 @@ else:
     ExponentialDecaySchedulerTensorFlow = ExponentialDecaySchedulerTensorFlow2
     InverseTimeDecaySchedulerTensorFlow = InverseTimeDecaySchedulerTensorFlow2
     ZarembaDecaySchedulerTensorFlow = ZarembaDecaySchedulerTensorFlow2
+    register(LinearDecaySchedulerTensorFlow, MEAD_LAYERS_LR_SCHEDULERS, "linear", "lr_scheduler")
 
 register(ConstantSchedulerTensorFlow, MEAD_LAYERS_LR_SCHEDULERS, "default", "lr_scheduler")
 register(WarmupLinearSchedulerTensorFlow, MEAD_LAYERS_LR_SCHEDULERS, "warmup_linear", "lr_scheduler")
