@@ -1,6 +1,15 @@
 import numpy as np
 from typing import Dict, List
-from eight_mile.tf.layers import TransformerEncoderStack, TransformerEncoder, TransformerDecoderStack, TransformerDecoder, MultiHeadedAttention, FFN
+from eight_mile.tf.layers import (
+    TransformerEncoderStack,
+    TransformerEncoder,
+    TransformerDecoderStack,
+    TransformerDecoder,
+    MultiHeadedAttention,
+    PassThru,
+    FFN,
+    SingleHeadReduction
+)
 from eight_mile.tf.embeddings import LookupTableEmbeddings, LearnedPositionalLookupTableEmbeddingsWithBias, LearnedPositionalLookupTableEmbeddings
 
 import tensorflow as tf
@@ -146,7 +155,7 @@ def to_decoder_array(tf_decoder: TransformerDecoder, name: str) -> Dict:
     d = {}
     d.update(to_weight_array(tf_decoder.ln1, f"{name}/ln1"))
     d.update(to_weight_array(tf_decoder.ln2, f"{name}/ln2"))
-    d.update(to_weight_array(tf_decoder.ln2, f"{name}/ln3"))
+    d.update(to_weight_array(tf_decoder.ln3, f"{name}/ln3"))
     d.update(to_attn_array(tf_decoder.self_attn, f"{name}/self_attn"))
     d.update(to_attn_array(tf_decoder.src_attn, f"{name}/src_attn"))
     d.update(to_ffn_array(tf_decoder.ffn, f"{name}/ffn"))
@@ -234,7 +243,7 @@ def from_encoder_stack_array(tf_encoder_stack: TransformerEncoderStack, d: Dict,
 def to_decoder_stack_array(
     tf_decoder_stack: TransformerDecoderStack, name: str = "TransformerDecoderStack"
 ) -> Dict:
-    """Convert a `TransformerEncoderStack` to a set of weigths
+    """Convert a `TransformerEncoderStack` to a set of weights
 
     :param tf_encoder_stack: A transformer encoder stack
     :param name: A name
@@ -247,6 +256,36 @@ def to_decoder_stack_array(
         d.update(to_decoder_array(dec_tf, f"{name}/{i}"))
     return d
 
+def from_decoder_array(tf_decoder: TransformerDecoder, d: Dict, name: str):
+    """Restore a `TransformerDecoder` layer from a set of numpy arrays
+
+    :param tf_decoder: A `TransformerDecoder` layer
+    :param d: A Dict of arrays by key
+    :param name: The layer name
+    :return: None
+    """
+    from_weight_array(tf_decoder.ln1, d, f"{name}/ln1")
+    from_weight_array(tf_decoder.ln2, d, f"{name}/ln2")
+    from_weight_array(tf_decoder.ln3, d, f"{name}/ln3")
+    from_attn_array(tf_decoder.src_attn, d, f"{name}/src_attn")
+    from_attn_array(tf_decoder.self_attn, d, f"{name}/self_attn")
+    from_ffn_array(tf_decoder.ffn, d, f"{name}/ffn")
+
+def from_decoder_stack_array(
+        tf_decoder_stack: TransformerDecoderStack, d: Dict, name: str = "TransformerDecoderStack"
+):
+    """Restore weights from a `TransformerDecoderStack`
+
+    :param pytorch_decoder_stack: A transformer decoder stack
+    :param d: A Dict containing sets of arrays
+    :param name: A name for this primitive
+    :return: None
+    """
+    if isinstance(tf_decoder_stack.ln, tf.keras.layers.LayerNormalization):
+        from_weight_array(tf_decoder_stack.ln, d, f"{name}/ln")
+    for i, dec_pyt in enumerate(tf_decoder_stack.decoders):
+        from_decoder_array(dec_pyt, d, f"{name}/{i}")
+
 
 def to_tlm_array(tf_tlm: tf.keras.layers.Layer, embeddings_keys: List[str] = None, name: str = "TLM") -> Dict:
     """Convert a Transformer LM-type module to a set of weights in a Dict
@@ -312,6 +351,26 @@ def save_tlm_npz(tf_tlm: tf.keras.layers.Layer, npz: str, embeddings_keys: List[
     np.savez(npz, **d)
 
 
+def save_tlm_output_npz(tlm: tf.keras.layers.Layer, npz: str, embeddings_keys: List[str] = None, name: str = "TLM", verbose: bool = False):
+    """Save a TLM to an NPZ file with an output layer
+
+    :param tlm: A Transformer LM-type module
+    :param npz: A file to save
+    :param embeddings_keys: A key to get embeddings from.  Defaults to `None`, in which case, all embeddings are written
+    :param name: A name for this TLM
+    :param verbose: whether output
+    :return: None
+    """
+    d = to_tlm_array(tlm, embeddings_keys, name)
+    if hasattr(tlm, 'output'):
+        d.update(to_weight_array(tlm.output, name=f'{name}/output'))
+    elif hasattr(tlm, 'output_layer'):
+        d.update(to_weight_array(tlm.output_layer, name=f'{name}/output'))
+    if verbose:
+        print(d.keys())
+    np.savez(npz, **d)
+
+
 def save_transformer_seq2seq_npz(tf_seq2seq: tf.keras.layers.Layer, npz: str, src_embeddings_keys: List[str] = None,
                                  tgt_embedding_key: str = 'y', name: str = "Seq2Seq", verbose: bool = False):
     """Save a Transformer seq2seq file out
@@ -343,6 +402,157 @@ def save_transformer_seq2seq_npz(tf_seq2seq: tf.keras.layers.Layer, npz: str, sr
     np.savez(npz, **enc, **dec)
 
 
+def load_transformer_seq2seq_npz(tf_seq2seq: tf.keras.layers.Layer,
+                                 npz: str, src_embeddings_keys: List[str] = None,
+                                 tgt_embedding_key: str = 'y', name: str = "Seq2Seq"):
+    """Save a Transformer seq2seq file out
+
+    The will be in pytorch_seq2seq.encoder.transformer, and the usual conversions work for that (via `to_tlm_array()`).
+    The decoder requires a new converter for the portion containing attention weights between the encoder and the decoder
+
+    :param pytorch_seq2seq: A Transformer Seq2Seq module
+    :param npz: The file name
+    :param src_embeddings_keys: An optional list of the src embeddings keys to load, otherwise use what we find
+    :param tgt_embedding_key: An optional tgt embedding, otherwise assume 'y' (TODO: bad assumption?)
+    :param name: An optional name of the model in the NPZ, otherwise assume `Seq2Seq`
+    """
+
+    d = np.load(npz)
+
+    transformer = tf_seq2seq.encoder.transformer
+    from_encoder_stack_array(transformer, d, name=f"{name}/TransformerEncoderStack")
+
+    enc_keys_to_restore = src_embeddings_keys if src_embeddings_keys else list(tf_seq2seq.src_embeddings.keys())
+
+    for embeddings_key in enc_keys_to_restore:
+        from_embed_array(tf_seq2seq.src_embeddings[embeddings_key], d, f"{name}/SrcEmbeddings/{embeddings_key}")
+
+    transformer_decoder = tf_seq2seq.decoder.transformer_decoder
+
+    from_decoder_stack_array(transformer_decoder, d,  name=f"{name}/TransformerDecoderStack")
+    from_embed_array(tf_seq2seq.decoder.tgt_embeddings, d, name=f"{name}/TgtEmbedding/{tgt_embedding_key}")
+
+
+def to_attn_pool_array(tf_attn_pool: tf.keras.layers.Layer, name: str) -> Dict:
+    """Convert a self-attention module to a set of arrays
+
+    :param tf_attn_pool: The attention pooling layer of a dual encoder, this could be single head or 2-headed
+    :return: A Dict containing the arrays by key
+    """
+    d = {}
+
+    if isinstance(tf_attn_pool, SingleHeadReduction):
+        d.update(to_weight_array(tf_attn_pool.w_Q, f"{name}/w_Q"))
+        d.update(to_weight_array(tf_attn_pool.w_K, f"{name}/w_K"))
+    else:
+        d.update(to_weight_array(tf_attn_pool.reduction1.w_Q, f"{name}/reduction1/w_Q"))
+        d.update(to_weight_array(tf_attn_pool.reduction1.w_K, f"{name}/reduction1/w_K"))
+
+        d.update(to_weight_array(tf_attn_pool.reduction2.w_Q, f"{name}/reduction2/w_Q"))
+        d.update(to_weight_array(tf_attn_pool.reduction2.w_K, f"{name}/reduction2/w_K"))
+    return d
+
+
+def from_attn_pool_array(tf_attn_pool: tf.keras.layers.Layer, d: Dict, name: str):
+    """Restore a self-attention pooling module from a set of keys
+
+    :param tf_attn_pool: The attention pooling layer of a dual encoder, this could be single head or 2-headed
+    :param d: A Dict of arrays by key
+    :param name: The name of the layer
+    """
+
+    if isinstance(tf_attn_pool, SingleHeadReduction):
+        from_weight_array(tf_attn_pool.w_Q, d, f"{name}/w_Q")
+        from_weight_array(tf_attn_pool.w_K, d, f"{name}/w_K")
+
+    else:
+        from_weight_array(tf_attn_pool.reduction1.w_Q, d, f"{name}/reduction1/w_Q")
+        from_weight_array(tf_attn_pool.reduction1.w_K, d, f"{name}/reduction1/w_K")
+        from_weight_array(tf_attn_pool.reduction2.w_Q, d, f"{name}/reduction2/w_Q")
+        from_weight_array(tf_attn_pool.reduction2.w_K, d, f"{name}/reduction2/w_K")
+
+
+def save_transformer_de_npz(tf_de: tf.keras.layers.Layer, npz: str, embeddings_keys: List[str] = None,
+                            name: str = "TLM", verbose: bool = False):
+    """Save a Transformer de file out
+
+    A Dual-Encoder will have 2 transformer layers with shared weights.  Because of this, when we save we only
+    want to save the first layer.  However, the upper "stacking" layers may be different
+
+    The encoder will have a transformer stack followed optionally by single or dual headed attention, and finally
+    either a linear or FFN stack of layers
+
+    :param tf_de: A Transformer Dual Encoder module
+    """
+
+    enc = {}
+    transformer = tf_de.transformer
+    enc.update(to_encoder_stack_array(transformer, name=f"{name}/TransformerEncoderStack"))
+    enc_keys_to_write = embeddings_keys if embeddings_keys else list(tf_de.embeddings.keys())
+
+    for embeddings_key in enc_keys_to_write:
+        enc.update(to_embed_array(tf_de.embeddings[embeddings_key], name=f"{name}/Embeddings/{embeddings_key}"))
+
+    enc.update(to_attn_pool_array(tf_de.reduction_layer, f"{name}/ReductionLayer"))
+
+    ff1 = tf_de.ff1
+    if isinstance(ff1, tf.keras.layers.Dense):
+        enc.update(to_weight_array(ff1, f"{name}/ff1"))
+    elif not isinstance(ff1, PassThru):
+        raise Exception("We dont currently support stacking layers in dual-encoder serialization")
+    ff2 = tf_de.ff2
+    if isinstance(ff2, tf.keras.layers.Dense):
+        enc.update(to_weight_array(ff2, f"{name}/ff2"))
+    elif not isinstance(ff2, PassThru):
+        raise Exception("We dont currently support stacking layers in dual-encoder serialization")
+
+    np.savez(npz, **enc)
+
+
+def load_transformer_de_npz(tf_de: tf.keras.layers.Layer,
+                            npz: str, embeddings_keys: List[str] = None,
+                            name: str = "TLM"):
+    """Load a dual-encoder from NPZ
+
+    A Dual-Encoder will have 2 transformer layers with shared weights.  Because of this, when we save we only
+    want to save the first layer.  However, the upper "stacking" layers may be different
+
+    The encoder will have a transformer stack followed optionally by single or dual headed attention, and finally
+    either a linear or FFN stack of layers
+
+    :param tf_de: A Transformer Dual Encoder module
+    :param npz:
+    :param embeddings_keys:
+    :param name:
+    :return:
+    """
+
+    d = np.load(npz)
+
+    transformer = tf_de.encoder.transformer
+    from_encoder_stack_array(transformer, d, name=f"{name}/TransformerEncoderStack")
+
+    enc_keys_to_restore = embeddings_keys if embeddings_keys else list(tf_de.embeddings.keys())
+
+    for embeddings_key in enc_keys_to_restore:
+        from_embed_array(tf_de.embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}")
+
+    from_attn_pool_array(tf_de.reduction, d, name=f"{name}/ReductionLayer")
+
+    if hasattr(tf_de, 'ff1'):
+        ff1 = tf_de.ff1
+        if isinstance(ff1, tf.keras.layers.Dense):
+            from_weight_array(ff1, d, f"{name}/ff1")
+        elif not isinstance(ff1, PassThru):
+            raise Exception("We dont currently support stacking layers in dual-encoder serialization")
+    if hasattr(tf_de, 'ff2'):
+        ff2 = tf_de.ff2
+        if isinstance(ff2, tf.keras.layers.Dense):
+            from_weight_array(ff2, d, f"{name}/ff2")
+        elif not isinstance(ff2, PassThru):
+            raise Exception("We dont currently support stacking layers in dual-encoder serialization")
+
+
 def load_tlm_npz(tf_tlm: tf.keras.layers.Layer, npz: str, embeddings_key: str = 'x', name: str = "TLM"):
     """Restore a TLM-like model (possibly a `Model` for fine-tuning
 
