@@ -2,6 +2,7 @@ import re
 import io
 import os
 import sys
+import time
 import json
 import logging
 import inspect
@@ -86,12 +87,46 @@ def register(cls, registry, name=None, error=""):
     return cls
 
 
+class classproperty(property):
+    def __get__(self, cls, owner):
+        return classmethod(self.fget).__get__(None, owner)()
+
+
 @export
 class Offsets:
-    """Support pre 3.4"""
+    INDICES = {
+        "PAD": 0,
+        "GO": 1,
+        "EOS": 2,
+        "UNK": 3,
+        "OFFSET": 4
+    }
+    VALUES = [
+        "<PAD>", "<GO>", "<EOS>", "<UNK>"
+    ]
+
+    @classproperty
+    def PAD(cls):
+        return cls.INDICES["PAD"]
+
+    @classproperty
+    def GO(cls):
+        return cls.INDICES["GO"]
+
+    @classproperty
+    def EOS(cls):
+        return cls.INDICES["EOS"]
+
+    @classproperty
+    def UNK(cls):
+        return cls.INDICES["UNK"]
+
+    @classproperty
+    def OFFSET(cls):
+        return cls.INDICES["OFFSET"]
+
+
 
-    PAD, GO, EOS, UNK, OFFSET = range(0, 5)
-    VALUES = ["<PAD>", "<GO>", "<EOS>", "<UNK>"]
 
 
 @export
@@ -430,6 +465,9 @@ def validate_url(url: str) -> bool:
     )
     return re.match(regex, url) is not None
 
+@export
+def pads(shape, dtype):
+    return np.full(shape, Offsets.PAD, dtype)
 
 
 @export
@@ -461,7 +499,7 @@ def read_config_stream(config_stream) -> Dict:
             path_to_save, _ = urlretrieve(config_stream)
             return read_config_stream(path_to_save)
         else:
-            logger.info("No file found '{}...', loading as string".format(config_stream[:12]))
+            logger.info("No file found '...{}', loading as string".format(config_stream[-32:]))
     return json.loads(config)
 
 
@@ -1454,8 +1492,9 @@ def f_score(overlap_count: int, gold_count: int, guess_count: int, f: int = 1) -
 
 
 @export
-def get_env_gpus() -> List[str]:
-    return os.getenv("CUDA_VISIBLE_DEVICES", os.getenv("NVIDIA_VISIBLE_DEVICES", os.getenv("NV_GPU", "0"))).split(",")
+def get_env_gpus(backoff="0") -> List[str]:
+    list_or_none = os.getenv("CUDA_VISIBLE_DEVICES", os.getenv("NVIDIA_VISIBLE_DEVICES", os.getenv("NV_GPU", backoff)))
+    return [] if list_or_none is None else list_or_none.split(",")
 
 
 @export
@@ -1605,8 +1644,8 @@ def to_numpy(x):
 def mlm_masking(inputs, mask_value, vocab_size, ignore_prefix, ignore_suffix, pad_y=True):
     labels = np.copy(inputs)
     masked_indices = np.random.binomial(size=len(inputs), n=1, p=0.15)
-    masked_indices = masked_indices & (labels != 0)
-    masked_indices[np.random.randint(1, sum(labels != 0))] = 1  #ensure at least one token is masked
+    masked_indices = masked_indices & (labels != Offsets.PAD)
+    masked_indices[np.random.randint(1, sum(labels != Offsets.PAD))] = 1  #ensure at least one token is masked
     if ignore_prefix:
         masked_indices[0] = 0
     if ignore_suffix:
@@ -1683,6 +1722,51 @@ class Average:
         return fmtstr.format(**self.__dict__)
 
 
+@export
+class Timer:
+
+    def __init__(self):
+        """Start the timer
+        """
+        self._started = 0
+        self.start()
+
+    def start(self):
+        """Start the timer.  Any old value overwritten
+        :return:
+        """
+        self._started = time.perf_counter()
+
+    def elapsed(self, return_minutes=False):
+        """Return the elapsed time (s) since the timer was started, w/o resetting start time
+
+        :param return_minutes if True, return minutes
+        :return: elapsed
+        """
+        elapsed = time.perf_counter() - self._started
+        return (elapsed / 60) if return_minutes else elapsed
+
+    def stop(self, return_minutes=False):
+        """Stop and null out the timer, returning the elapsed time (s) between when the timer was started and this call
+
+        :param return_minutes if True, return minutes
+        :return: elapsed
+        """
+        elapsed = self.elapsed()
+        self._started = 0
+        return (elapsed / 60) if return_minutes else elapsed
+
+    def restart(self, return_minutes=False):
+        """Return the elapsed time since the last start call and reset the timer to the current time
+
+        :param return_minutes if True, return minutes
+        :return: elapsed prior to restart
+        """
+        old_start = self._started
+        self.start()
+        elapsed = self._started - old_start
+        return (elapsed / 60) if return_minutes else elapsed
+
 @export
 def undo_bpe(seq: str) -> str:
     """Undo the BPE splits to make Bleu comparable.
